"""A parallel ODE solver for PyTorch"""

__version__ = "0.1.1.post1"

from .adjoints import AutoDiffAdjoint, BacksolveAdjoint, JointBacksolveAdjoint
from .interface import register_method, solve_ivp
from .problems import InitialValueProblem
from .single_step_methods import Dopri5, Euler, Heun, Tsit5
from .solution import Solution
from .status_codes import Status
from .step_size_controllers import (
    FixedStepController,
    IntegralController,
    PIDController,
)
from .terms import ODETerm

register_method("heun", Heun)
register_method("dopri5", Dopri5)
register_method("tsit5", Tsit5)



from typing import Optional

from .typing import *


class InitialValueProblem:
    """An initial value problem.

    You can have different data types in the data and time domains. If your start, end
    and evaluation points are in double precision, all time computations will be done in
    double precision, while all data computations will happen in the data dtype. Note,
    that this requires your term to deal with inputs of different data types and still
    return the correct dtype (the one of `y0`).

    If you do not set explicit evaluation points, the solution will only be evaluated at
    the end points `t_end`. This improves performance of the solver loop in cases where
    we only care about the solution at the end point such as continuous normalizing
    flows.
    """

    def __init__(
        self,
        y0: DataTensor,
        t_start: Optional[TimeTensor] = None,
        t_end: Optional[TimeTensor] = None,
        t_eval: Optional[EvaluationTimesTensor] = None,
    ):
        self.y0 = y0
        self.t_eval = t_eval

        if t_start is None:
            assert t_eval is not None
            t_start = t_eval[:, 0]
        self.t_start = t_start

        if t_end is None:
            assert t_eval is not None
            t_end = t_eval[:, -1]
        self.t_end = t_end


        self.time_direction = torch.where(self.t_end > self.t_start, 1, -1)

        if not torch.jit.is_scripting():
            assert y0.ndim == 2
            assert self.t_start.ndim == 1
            assert self.t_end.ndim == 1
            assert same_dtype(self.t_start, self.t_end)
            assert same_shape(y0, self.t_start, self.t_end, dim=0)
            assert same_device(y0, self.t_start, self.t_end)

            if t_eval is not None:
                assert t_eval.ndim == 2
                assert same_dtype(self.t_start, t_eval)
                assert same_shape(self.t_start, t_eval, dim=0)
                assert same_device(self.t_start, t_eval)

    @property
    def data_dtype(self):
        return self.y0.dtype

    @property
    def time_dtype(self):
        return self.t_start.dtype

    @property
    def device(self):
        return self.y0.device

    @property
    def batch_size(self):
        return self.y0.shape[0]

    @property
    def n_features(self):
        return self.y0.shape[1]

    @property
    def n_evaluation_points(self):
        t_eval = self.t_eval
        if t_eval is None:
            return 0
        else:
            return t_eval.shape[1]

    def __repr__(self):
        return (
            f"InitialValueProblem(y0={self.y0}, t_start={self.t_start}, "
            f"t_end={self.t_end}, t_eval={self.t_eval})"
        )



from typing import Optional

import torch
from torchtyping import TensorType, is_float


def same_dtype(*tensors: torch.Tensor):
    if len(tensors) <= 1:
        return True
    for a, b in zip(tensors[:-1], tensors[1:]):
        if a.dtype != b.dtype:
            return False
    return True


def same_device(*tensors: torch.Tensor):
    if len(tensors) <= 1:
        return True
    for a, b in zip(tensors[:-1], tensors[1:]):
        if a.device != b.device:
            return False
    return True


def same_shape(*tensors: torch.Tensor, dim: Optional[int] = None):
    if len(tensors) <= 1:
        return True
    for a, b in zip(tensors[:-1], tensors[1:]):
        if dim is None:
            if a.shape != b.shape:
                return False
        else:
            if a.shape[dim] != b.shape[dim]:
                return False
    return True


################
# Tensor Types #
################

# These are subclasses of TensorType instead of using TensorType annotations directly,
# because TorchScript does not support custom type constructors. In this way, we can
# continue to document the shapes and types of tensors while being TorchScript
# compatible, see [1].
#
# [1] https://github.com/patrick-kidger/torchtyping/issues/13


class DataTensor(TensorType["batch", "feature", is_float]):
    pass


class NormTensor(TensorType["batch", is_float]):
    pass


class SolutionDataTensor(TensorType["batch", "time", "feature", is_float]):
    pass


class TimeTensor(TensorType["batch", is_float]):
    pass


class EvaluationTimesTensor(TensorType["batch", "time", is_float]):
    pass


class AcceptTensor(TensorType["batch", torch.bool]):
    pass


class StatusTensor(TensorType["batch", torch.long]):
    pass


class InterpTimeTensor(TensorType["interp-points", is_float]):
    pass


class InterpDataTensor(TensorType["interp-points", "feature", is_float]):
    pass


class SampleIndexTensor(TensorType["interp-points", torch.long]):
    pass



from typing import Optional

import torch

from ..interpolation import FourthOrderPolynomialInterpolation
from ..terms import ODETerm
from .runge_kutta import ButcherTableau, ERKInterpolationData, ExplicitRungeKutta


class Dopri5(ExplicitRungeKutta):
    TABLEAU = ButcherTableau.from_lists(
        c=[0.0, 1 / 5, 3 / 10, 4 / 5, 8 / 9, 1.0, 1.0],
        a=[
            [],
            [1 / 5],
            [3 / 40, 9 / 40],
            [44 / 45, -56 / 15, 32 / 9],
            [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729],
            [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656],
            [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84],
        ],
        b=[35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84, 0.0],
        b_low_order=[
            1951 / 21600,
            0,
            22642 / 50085,
            451 / 720,
            -12231 / 42400,
            649 / 6300,
            1 / 60,
        ],
        b_other=[
            # Coefficients for y at the mid point
            [
                6025192743 / 30085553152 / 2,
                0,
                51252292925 / 65400821598 / 2,
                -2691868925 / 45128329728 / 2,
                187940372067 / 1594534317056 / 2,
                -1776094331 / 19743644256 / 2,
                11237099 / 235043384 / 2,
            ]
        ],
    )

    def __init__(self, term: Optional[ODETerm] = None):
        super().__init__(term, Dopri5.TABLEAU)

    @torch.jit.export
    def convergence_order(self):
        return 5

    @torch.jit.export
    def build_interpolation(self, data: ERKInterpolationData):
        b_other = data.tableau.b_other
        assert b_other is not None
        b_mid = b_other[0]
        return FourthOrderPolynomialInterpolation.from_k(
            data.t0, data.dt, data.y0, data.y1, data.k, b_mid
        )



from typing import Any, Optional

import functorch
import torch
import torch.nn as nn

from .problems import InitialValueProblem
from .single_step_methods import SingleStepMethod
from .solution import Solution
from .status_codes import Status
from .step_size_controllers import StepSizeController
from .terms import ODETerm
from .typing import *


class Adjoint:
    def solve(
        self,
        problem: InitialValueProblem,
        term: Optional[ODETerm] = None,
        dt0: Optional[TimeTensor] = None,
        args: Any = None,
    ) -> Solution:
        raise NotImplementedError()


class AutoDiffAdjoint(nn.Module):
    def __init__(
        self,
        step_method: SingleStepMethod,
        step_size_controller: StepSizeController,
        *,
        max_steps: Optional[int] = None,
        backprop_through_step_size_control: bool = True,
    ):
        super().__init__()

        self.step_method = step_method
        self.step_size_controller = step_size_controller
        self.max_steps = max_steps
        self.backprop_through_step_size_control = backprop_through_step_size_control

    @torch.jit.export
    def solve(
        self,
        problem: InitialValueProblem,
        term: Optional[ODETerm] = None,
        dt0: Optional[TimeTensor] = None,
        args: Any = None,
    ) -> Solution:
        step_method, step_size_controller = self.step_method, self.step_size_controller
        device, batch_size = problem.device, problem.batch_size
        t_start, t_end, t_eval = problem.t_start, problem.t_end, problem.t_eval
        time_direction = problem.time_direction.to(dtype=t_start.dtype)

        ###############################
        # Initialize the solver state #
        ###############################

        t = t_start
        y = problem.y0
        stats_n_steps = y.new_zeros(batch_size, dtype=torch.long)
        stats_n_accepted = y.new_zeros(batch_size, dtype=torch.long)
        stats: Dict[str, Any] = {}

        # TorchScript is not smart enough yet to figure out that we only access these
        # variables when they have been defined, so we have to always define them and
        # intialize them to any valid tensor.
        y_eval: torch.Tensor = y
        not_yet_evaluated: torch.Tensor = y
        minus_t_eval_normalized: torch.Tensor = y

        if t_eval is not None:
            y_eval = y.new_empty(
                (batch_size, problem.n_evaluation_points, problem.n_features)
            )

            # Keep track of which evaluation points have not yet been handled
            not_yet_evaluated = torch.ones_like(t_eval, dtype=torch.bool)

        # Normalize the time direction of the evaluation and end times for faster
        # comparisons
        minus_t_end_normalized = -time_direction * t_end
        if t_eval is not None:
            minus_t_eval_normalized = -time_direction[:, None] * t_eval

        # Keep track of which solves are still running
        running = y.new_ones(batch_size, dtype=torch.bool)

        # Initialize additional statistics to track for the integration term
        term_ = term
        if torch.jit.is_scripting() or term is None:
            assert term is None, "The integration term is fixed for JIT compilation"
            term_ = self.step_method.term
        assert term_ is not None
        term_.init(problem, stats)

        # Compute an initial step size
        convergence_order = step_method.convergence_order()
        dt, controller_state, f0 = step_size_controller.init(
            term, problem, convergence_order, dt0, stats=stats, args=args
        )
        method_state = step_method.init(term, problem, f0, stats=stats, args=args)

        # TorchScript does not support set_grad_enabled, so we detach manually
        if not self.backprop_through_step_size_control:
            dt = dt.detach()

        ##############################################
        # Take care of evaluation exactly at t_start #
        ##############################################

        # We copy the initial state into the evaluation if the first evaluation point
        # happens to be exactly `t_start`. This is required so that we can later assume
        # that rejection of the step (and therefore also no change in `t`) means that we
        # also did not pass any evaluation points.
        if t_eval is not None:
            eval_at_start = t_eval[:, 0] == t_start
            y_eval[eval_at_start, 0] = y[eval_at_start]
            not_yet_evaluated[eval_at_start, 0] = False

        ####################################
        # Solve the initial value problems #
        ####################################

        # Iterate the single step method until all ODEs have been solved up to their end
        # point or any of them failed
        max_steps = self.max_steps
        while True:
            step_out = step_method.step(
                term, running, y, t, dt, method_state, stats=stats, args=args
            )
            step_result, interp_data, method_state_next, method_status = step_out
            controller_out = step_size_controller.adapt_step_size(
                t, dt, y, step_result, controller_state, stats
            )
            accept, dt_next, controller_state_next, controller_status = controller_out

            # TorchScript does not support set_grad_enabled, so we detach manually
            if not self.backprop_through_step_size_control:
                dt_next = dt_next.detach()

            # Update the solver state where the step was accepted
            to_update = accept & running
            t = torch.where(to_update, t + dt, t)
            y = torch.where(to_update[:, None], step_result.y, y)
            method_state = step_method.merge_states(
                to_update, method_state_next, method_state
            )

            #####################
            # Update statistics #
            #####################

            stats_n_steps.add_(running)
            stats_n_accepted.add_(to_update)

            ##################################
            # Update solver state and status #
            ##################################

            # Stop a solve if `t` has passed its endpoint in the direction of time
            running = torch.addcmul(minus_t_end_normalized, time_direction, t) < 0.0

            status = method_status
            if status is None:
                status = controller_status
            elif controller_status is not None:
                status = torch.maximum(status, controller_status)
            if max_steps is not None:
                status = torch.where(
                    stats_n_steps >= max_steps,
                    Status.REACHED_MAX_STEPS.value,
                    status if status is not None else Status.SUCCESS.value,
                )

            # We evaluate the termination condition here already and initiate a
            # non-blocking transfer to the CPU to increase the chance that we won't have
            # to wait for the result when we actually check the termination condition
            continue_iterating = torch.any(running)
            if status is not None:
                continue_iterating = continue_iterating & torch.all(
                    status == Status.SUCCESS.value
                )
            continue_iterating = continue_iterating.to("cpu", non_blocking=True)

            # There is a bug as of pytorch 1.12.1 where non-blocking transfer from
            # device to host can sometimes gives the wrong result, so we place this
            # event after the transfer to ensure that the transfer has actually happened
            # by the time we evaluate the result.
            if device.type == "cuda":
                continue_iterating_done = torch.cuda.Event()
                continue_iterating_done.record(torch.cuda.current_stream(device))
            else:
                continue_iterating_done = None

            #########################
            # Evaluate the solution #
            #########################

            # Evaluate the solution at all evaluation points that have been passed in
            # this step.
            #
            # We always build the interpolation and evaluate it, even if no evaluation
            # points have actually been passed, because this avoids a CPU-GPU
            # synchronization and for time series models we expect that most steps will
            # pass at least one evaluation point across the whole batch (usually more).
            if t_eval is not None:
                to_be_evaluated = (
                    torch.addcmul(
                        minus_t_eval_normalized,
                        time_direction[:, None],
                        t[:, None],
                    )
                    > 0.0
                ) & not_yet_evaluated
                if to_be_evaluated.any():
                    interpolation = step_method.build_interpolation(interp_data)
                    nonzero = to_be_evaluated.nonzero()
                    sample_idx, eval_t_idx = nonzero[:, 0], nonzero[:, 1]
                    y_eval[sample_idx, eval_t_idx] = interpolation.evaluate(
                        t_eval[sample_idx, eval_t_idx], sample_idx
                    )

                    not_yet_evaluated = torch.logical_xor(
                        to_be_evaluated, not_yet_evaluated
                    )

            ########################
            # Update the step size #
            ########################

            # We update the step size and controller state only for solves which will
            # still be running in the next iteration. Otherwise, a finished instance
            # with an adaptive step size controller could reach infinite step size if
            # its final error was small and another instance is running for many steps.
            # This would then cancel the solve even though the "problematic" instance is
            # not even supposed to be running anymore.

            dt = torch.where(running, dt_next, dt)
            controller_state = step_size_controller.merge_states(
                running, controller_state_next, controller_state
            )

            if continue_iterating_done is not None:
                continue_iterating_done.synchronize()
            if continue_iterating:
                continue

            ##################################################
            # Finalize the solver and construct the solution #
            ##################################################

            # Ensure that the user always gets a status tensor
            if status is None:
                status = torch.tensor(
                    Status.SUCCESS.value, dtype=torch.long, device=device
                ).expand(batch_size)

            # Put the step statistics into the stats dict in the end, so that
            # we don't have to type-assert all the time in torchscript
            stats["n_steps"] = stats_n_steps
            stats["n_accepted"] = stats_n_accepted

            # The finalization scope is in the scope of the while loop so that the
            # `t_eval is None` case can access the `interp_data` in TorchScript.
            # Declaring `interp_data` outside of the loop does not work because its type
            # depends on the step method.

            if t_eval is not None:
                # Report the number of evaluation steps that have been initialized with
                # actual data at termination. Depending on the termination condition,
                # the data might be NaN or inf but it will not be uninitialized memory.
                #
                # As of torch 1.12.1, searchsorted is not implemented for bool tensors,
                # so we convert to int first.
                stats["n_initialized"] = torch.searchsorted(
                    not_yet_evaluated.int(),
                    torch.ones((batch_size, 1), dtype=torch.int, device=device),
                ).squeeze(dim=1)

                return Solution(ts=t_eval, ys=y_eval, stats=stats, status=status)
            else:
                # Evaluate the solution only at the end point, e.g. in continuous
                # normalizing flows
                interpolation = step_method.build_interpolation(interp_data)
                y_end = interpolation.evaluate(
                    t_end, torch.arange(batch_size, device=device)
                )

                stats["n_initialized"] = torch.ones(
                    batch_size, dtype=torch.long, device=device
                )

                return Solution(
                    ts=t_end[:, None], ys=y_end[:, None], stats=stats, status=status
                )

        assert False, "unreachable"

    def __repr__(self):
        return (
            f"AutoDiffAdjoint(step_method={self.step_method}, "
            f"step_size_controller={self.step_size_controller}, "
            f"max_steps={self.max_steps}, "
            f"backprop_through_step_size_control={self.backprop_through_step_size_control})"
        )


def flatten_tensors(*tensors):
    return [t.shape[1:] for t in tensors], torch.cat(
        [t.reshape((t.shape[0], -1)) for t in tensors], dim=1
    )


def unflatten_tensors(shapes, tensors):
    def prod(nums):
        p = 1
        for n in nums:
            p = p * n
        return p

    return [
        tensor.reshape((-1, *shape))
        for shape, tensor in zip(
            shapes, torch.split(tensors, [prod(shape) for shape in shapes], dim=1)
        )
    ]


class BacksolveFunction(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        solver,
        aug_solver,
        term,
        aug_term,
        y0,
        t_start,
        t_end,
        t_eval,
        args,
        *adjoint_params,
    ):
        ctx.aug_solver = aug_solver
        ctx.aug_term = aug_term
        ctx.args = args

        problem = InitialValueProblem(y0, t_start, t_end, t_eval)
        with torch.no_grad():
            solution = solver.solve(problem, dt0=None, args=args, term=term)

        ctx.stats = solution.stats
        ctx.save_for_backward(t_start, t_end, t_eval, solution.ys, *adjoint_params)

        return solution.ts, solution.ys, solution.stats, solution.status

    @staticmethod
    def backward(ctx, grad_ts, grad_ys, grad_stats, grad_status):
        aug_solver = ctx.aug_solver
        aug_term = ctx.aug_term
        args = ctx.args
        t_start, t_end, t_eval, ys, *adjoint_params = ctx.saved_tensors

        stats = ctx.stats
        stats.setdefault("backsolve", [])

        batch_size, n_eval, n_features = ys.shape
        shapes, aug_state = flatten_tensors(
            ys.new_zeros((batch_size, 1)),
            ys[:, -1],
            grad_ys[:, -1],
            *[torch.zeros_like(p).expand(batch_size, *p.shape) for p in adjoint_params],
        )
        with torch.no_grad():
            if t_eval is None:
                problem = InitialValueProblem(
                    aug_state, t_start=t_end, t_end=t_start, t_eval=None
                )
                solution = aug_solver.solve(
                    problem, dt0=None, args=(shapes, args), term=aug_term
                )
                aug_state = solution.ys[:, -1]

                stats["backsolve"].append(solution.stats)
            else:
                for i in range(n_eval - 1, 0, -1):
                    problem = InitialValueProblem(
                        aug_state,
                        t_start=t_eval[:, i],
                        t_end=t_eval[:, i - 1],
                        t_eval=None,
                    )
                    solution = aug_solver.solve(
                        problem, dt0=None, args=(shapes, args), term=aug_term
                    )
                    aug_state = solution.ys[:, -1]
                    aug_state[:, 1 : 1 + n_features] = ys[:, i - 1]
                    aug_state[:, 1 + n_features : 1 + 2 * n_features] += grad_ys[
                        :, i - 1
                    ]

                    stats["backsolve"].append(solution.stats)

        _, _, grad_y0, *grad_params = unflatten_tensors(shapes, aug_state)
        # Accumulate gradients over samples
        grad_params = [p.sum(dim=0) for p in grad_params]

        return None, None, None, None, grad_y0, None, None, None, None, *grad_params


class AugmentedDynamicsTerm(nn.Module):
    def __init__(
        self, term: ODETerm, vmap_args_dims=None, vmap_randomness: str = "error"
    ):
        super().__init__()

        self.term = term
        self.vmap_args_dims = vmap_args_dims
        self.vmap_randomness = vmap_randomness

        self.func, self.params, self.buffers = functorch.make_functional_with_buffers(
            self.term.f
        )

        def vjp_single_sample(t_i, y_i, adj_y_i, arg_i):
            def wrapper(params, t_, y_):
                if self.term.with_args:
                    return self.func(params, self.buffers, t_, y_, arg_i)
                else:
                    return self.func(params, self.buffers, t_, y_)

            dy, vjp = functorch.vjp(wrapper, self.params, t_i, y_i)
            vjp_params, vjp_t, vjp_y = vjp(-adj_y_i)
            return dy, vjp_t, vjp_y, vjp_params

        self.vjp_vf = functorch.vmap(
            vjp_single_sample,
            in_dims=(0, 0, 0, self.vmap_args_dims),
            randomness=self.vmap_randomness,
        )

    def init(self, problem: InitialValueProblem, stats: dict[str, Any]):
        return self.term.init(problem, stats)

    def vf(
        self, t: TimeTensor, y: DataTensor, stats: dict[str, Any], args: Any
    ) -> DataTensor:
        shapes, args = args

        adj_t, y, adj_y, *_ = unflatten_tensors(shapes, y)
        dy, vjp_t, vjp_y, vjp_params = self.vjp_vf(t, y, adj_y, args)

        return flatten_tensors(vjp_t[:, None], dy, vjp_y, *vjp_params)[1]


class BacksolveAdjoint(nn.Module):
    def __init__(
        self,
        term,
        step_method,
        step_size_controller,
        vmap_args_dims=None,
        vmap_randomness="error",
    ):
        super().__init__()

        self.term = term
        self.augmented_term = AugmentedDynamicsTerm(
            self.term, vmap_args_dims=vmap_args_dims, vmap_randomness=vmap_randomness
        )
        self.forward_adjoint = AutoDiffAdjoint(step_method, step_size_controller)
        self.backward_adjoint = AutoDiffAdjoint(step_method, step_size_controller)

    def solve(
        self,
        problem: InitialValueProblem,
        term: Optional[ODETerm] = None,
        dt0: Optional[TimeTensor] = None,
        args: Any = None,
    ) -> Solution:
        ts, ys, stats, status = BacksolveFunction.apply(
            self.forward_adjoint,
            self.backward_adjoint,
            self.term,
            self.augmented_term,
            problem.y0,
            problem.t_start,
            problem.t_end,
            problem.t_eval,
            args,
            *list(self.term.parameters()),
        )
        return Solution(ts, ys, stats, status)

    def __repr__(self):
        return (
            f"BacksolveAdjoint(term={self.term!r}, "
            f"augmented_term={self.augmented_term!r}, "
            f"forward_adjoint={self.forward_adjoint!r}, "
            f"backward_adjoint={self.backward_adjoint!r})"
        )


class UnwrappingODETerm(nn.Module):
    def __init__(self, term: ODETerm):
        super().__init__()

        self.term = term

    def init(self, problem: InitialValueProblem, stats: dict[str, Any]):
        return self.term.init(problem, stats)

    def vf(
        self, t: TimeTensor, y: DataTensor, stats: dict[str, Any], args: Any
    ) -> DataTensor:
        """Evaluate the vector field."""
        batch_size, t_intercept, t_slope, term_args = args

        # Map from the integration interval of the first instance onto the other
        # integration intervals
        t = torch.addcmul(t_intercept, t_slope, t)

        dy = self.term.vf(t, y[0].reshape((batch_size, -1)), stats, term_args)

        # Correct the y-derivative for the distortion through the linear mapping
        # according to the substitution rule
        dy = dy * t_slope[:, None]

        return dy.flatten()[None]


class JointAugmentedDynamicsTerm(nn.Module):
    def __init__(self, term: UnwrappingODETerm):
        super().__init__()

        self.term = term
        self.parameters = list(self.term.parameters())

    def init(self, problem: InitialValueProblem, stats: dict[str, Any]):
        return self.term.init(problem, stats)

    def vf(
        self, t: TimeTensor, y: DataTensor, stats: dict[str, Any], args: Any
    ) -> DataTensor:
        shapes, term_args = args

        adj_t, y, adj_y, *_ = unflatten_tensors(shapes, y)

        with torch.enable_grad():
            t_ = t.detach().requires_grad_()
            y_ = y.detach().requires_grad_()
            dy = self.term.vf(t_, y_, stats, term_args)
            vjp_t, vjp_y, *vjp_params = torch.autograd.grad(
                dy,
                [t_, y_] + self.parameters,
                -adj_y,
                allow_unused=True,
                retain_graph=True,
            )

        if vjp_t is None:
            vjp_t = torch.zeros_like(t)
        if vjp_y is None:
            vjp_y = torch.zeros_like(y)
        vjp_params = [
            vp if vp is not None else torch.zeros_like(p)
            for p, vp in zip(self.parameters, vjp_params)
        ]

        dy_joint = dy.flatten()[None]
        vjp_params_joint = [p[None] for p in vjp_params]
        return flatten_tensors(vjp_t[:, None], dy_joint, vjp_y, *vjp_params_joint)[1]


class JointBacksolveAdjoint(nn.Module):
    def __init__(self, term, step_method, step_size_controller):
        super().__init__()

        self.term = UnwrappingODETerm(term)
        self.augmented_term = JointAugmentedDynamicsTerm(self.term)
        self.forward_loop = AutoDiffAdjoint(step_method, step_size_controller)
        self.backward_loop = AutoDiffAdjoint(step_method, step_size_controller)

    def solve(
        self,
        problem: InitialValueProblem,
        term: Optional[ODETerm] = None,
        dt0: Optional[TimeTensor] = None,
        args: Any = None,
    ) -> Solution:
        y0 = problem.y0.flatten()[None]
        t_start = problem.t_start[:1]
        t_end = problem.t_end[:1]
        t_eval = problem.t_eval
        if t_eval is not None:
            eval_steps = torch.diff(t_eval, dim=1)
            normalized_steps = eval_steps / torch.maximum(
                eval_steps[:, :1], t_eval.new_tensor(1e-8)
            )
            assert (normalized_steps - normalized_steps[0]).abs().max() < 1e-8, (
                "JointBacksolveAdjoint can only be applied if all instances in "
                "the batch are evaluated at the same points in time"
            )
            t_eval = t_eval[:1]

        # Compute linear maps from the integration interval (and evaluation points) of
        # any instance onto the integration interval of the first instance in the batch
        all_slopes = problem.t_end - problem.t_start
        t_slope = all_slopes / all_slopes[0]
        t_intercept = problem.t_start - t_slope * problem.t_start[0]

        _, ys, stats, status = BacksolveFunction.apply(
            self.forward_loop,
            self.backward_loop,
            self.term,
            self.augmented_term,
            y0,
            t_start,
            t_end,
            t_eval,
            (problem.batch_size, t_intercept, t_slope, args),
            *list(self.term.parameters()),
        )

        ys = ys[0].unflatten(dim=1, sizes=(problem.batch_size, -1)).transpose(1, 0)
        if problem.t_eval is None:
            ts = problem.t_end
        else:
            ts = problem.t_eval
        return Solution(ts, ys, stats, status)

    def __repr__(self):
        return (
            f"JointBacksolveAdjoint(term={self.term!r}, "
            f"augmented_term={self.augmented_term!r}, "
            f"forward_loop={self.forward_loop!r}, "
            f"backward_loop={self.backward_loop!r})"
        )



from typing import Any, NamedTuple, Optional

import torch
import torch.nn as nn
from torchtyping import TensorType

from ..interpolation import LocalInterpolation
from ..problems import InitialValueProblem
from ..status_codes import Status
from ..terms import ODETerm
from ..typing import *
from .base import StepResult


class CoefficientVector(TensorType["nodes"]):
    pass


class RungeKuttaMatrix(TensorType["nodes", "weights"]):
    pass


class WeightVector(TensorType["weights"]):
    pass


class WeightMatrix(TensorType["rows", "weights"]):
    pass


class ButcherTableau:
    def __init__(
        self,
        # Coefficients for the evaluation nodes in time
        c: CoefficientVector,
        # Runge-Kutta matrix
        a: RungeKuttaMatrix,
        # Coefficients for the high-order solution estimate
        b: WeightVector,
        # Coefficients for the error estimate
        b_err: WeightVector,
        # Additional additional rows of the b matrix
        b_other: Optional[WeightMatrix] = None,
        fsal: Optional[bool] = None,
        ssal: Optional[bool] = None,
    ):
        self.c = c
        self.a = a
        self.b = b
        self.b_err = b_err
        self.b_other = b_other

        if fsal is None:
            fsal = self.is_fsal()
        self.fsal = fsal
        if ssal is None:
            ssal = self.is_ssal()
        self.ssal = ssal

    @staticmethod
    def from_lists(
        *,
        c: list[float],
        a: list[list[float]],
        b: list[float],
        b_err: Optional[list[float]] = None,
        b_low_order: Optional[list[float]] = None,
        b_other: Optional[list[list[float]]] = None,
    ):
        assert b_err is not None or b_low_order is not None, (
            "You have to provide either the weights for the error approximation"
            " or the weights of an embedded lower-order method"
        )

        n_nodes = len(c)
        n_weights = len(b)
        assert n_nodes == n_weights
        assert len(a) == n_nodes

        # Fill a up into a full square matrix
        a_full = [row + [0.0] * (n_weights - len(row)) for row in a]

        b_coeffs = torch.tensor(b, dtype=torch.float64)
        if b_err is None:
            assert b_low_order is not None
            assert len(b_low_order) == n_weights
            b_low_coeffs = torch.tensor(b_low_order, dtype=torch.float64)
            b_err_coeffs = b_coeffs - b_low_coeffs
        else:
            b_err_coeffs = torch.tensor(b_err, dtype=torch.float64)

        if b_other is None:
            b_other_coeffs = None
        else:
            b_other_coeffs = torch.tensor(b_other, dtype=torch.float64)
            assert b_other_coeffs.ndim == 2
            assert b_other_coeffs.shape[1] == n_weights

        return ButcherTableau(
            c=torch.tensor(c, dtype=torch.float64),
            a=torch.tensor(a_full, dtype=torch.float64),
            b=b_coeffs,
            b_err=b_err_coeffs,
            b_other=b_other_coeffs,
        )

    def to(
        self, device: torch.device, time_dtype: torch.dtype, data_dtype: torch.dtype
    ) -> "ButcherTableau":
        b_other = self.b_other
        if b_other is not None:
            b_other = b_other.to(device, data_dtype)
        return ButcherTableau(
            c=self.c.to(device, time_dtype),
            a=self.a.to(device, data_dtype),
            b=self.b.to(device, data_dtype),
            b_err=self.b_err.to(device, data_dtype),
            b_other=b_other,
            fsal=self.fsal,
            ssal=self.ssal,
        )

    @property
    def n_stages(self):
        return self.c.shape[0]

    def is_fsal(self):
        """Is `f(y0)` equal to `f(y1)` from the previous step?

        If that is the case, we can reuse the result from the previous step.
        """
        is_lower_triangular = (torch.triu(self.a, diagonal=1) == 0.0).all().item()
        first_node_is_t0 = (self.c[0] == 0.0).item()
        last_node_is_t1 = (self.c[-1] == 1.0).item()
        first_stage_explicit = (self.a[0, 0] == 0.0).item()
        return (
            is_lower_triangular
            and (self.b == self.a[-1]).all().item()
            and first_node_is_t0
            and last_node_is_t1
            and first_stage_explicit
        )

    def is_ssal(self):
        """Is the solution equal to the last stage result?

        If that is the case, we can avoid the final computation of the solution and
        return the last stage result instead.
        """
        is_lower_triangular = (torch.triu(self.a, diagonal=1) == 0.0).all().item()
        last_node_is_t1 = (self.c[-1] == 1.0).item()
        last_stage_explicit = (self.a[-1, -1] == 0.0).item()
        return (
            is_lower_triangular
            and (self.b == self.a[-1]).all().item()
            and last_node_is_t1
            and last_stage_explicit
        )


class ERKInterpolationData(NamedTuple):
    tableau: ButcherTableau
    t0: TimeTensor
    dt: TimeTensor
    y0: DataTensor
    y1: DataTensor
    k: torch.Tensor


class ERKState(NamedTuple):
    tableau: ButcherTableau
    prev_vf1: Optional[DataTensor]


class ExplicitRungeKutta(nn.Module):
    def __init__(self, term: Optional[ODETerm], tableau: ButcherTableau):
        super().__init__()

        self.term = term
        self.tableau = tableau

    @torch.jit.export
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        f0: Optional[DataTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> ERKState:
        if self.tableau.fsal:
            term_ = term
            if torch.jit.is_scripting() or term_ is None:
                assert term is None, "The integration term is fixed for JIT compilation"
                term_ = self.term
            assert term_ is not None

            if f0 is None:
                prev_vf1 = term_.vf(problem.t_start, problem.y0, stats, args)
            else:
                prev_vf1 = f0
        else:
            prev_vf1 = None

        return ERKState(
            tableau=self.tableau.to(
                device=problem.device,
                data_dtype=problem.data_dtype,
                time_dtype=problem.time_dtype,
            ),
            prev_vf1=prev_vf1,
        )

    @torch.jit.export
    def merge_states(self, accept: AcceptTensor, current: ERKState, previous: ERKState):
        prev_vf1 = previous.prev_vf1
        current_vf1 = current.prev_vf1
        if current_vf1 is None or prev_vf1 is None:
            return current
        else:
            return ERKState(
                current.tableau, torch.where(accept[:, None], current_vf1, prev_vf1)
            )

    @torch.jit.export
    def step(
        self,
        term: Optional[ODETerm],
        running: AcceptTensor,
        y0: DataTensor,
        t0: TimeTensor,
        dt: TimeTensor,
        state: ERKState,
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[StepResult, ERKInterpolationData, ERKState, Optional[StatusTensor]]:
        term_ = term
        if torch.jit.is_scripting() or term_ is None:
            assert term is None, "The integration term is fixed for JIT compilation"
            term_ = self.term
        assert term_ is not None
        tableau = state.tableau

        # Convert dt into the data dtype for dtype stability
        dt_data = dt.to(dtype=y0.dtype)

        prev_vf1 = state.prev_vf1
        vf0 = (
            prev_vf1
            if tableau.fsal and prev_vf1 is not None
            else term_.vf(t0, y0, stats, args)
        )
        y_i = y0
        k = vf0.new_empty((tableau.n_stages, vf0.shape[0], vf0.shape[1]))
        k[0] = vf0
        a = tableau.a
        t_nodes = torch.addcmul(t0, tableau.c[:, None], dt)
        for i in range(1, tableau.n_stages):
            y_i = torch.einsum("j, jbf -> bf", a[i, :i], k[:i])
            y_i = torch.addcmul(y0, dt_data[:, None], y_i)
            k[i] = term_.vf(t_nodes[i], y_i, stats, args)

        if tableau.ssal:
            y1 = y_i
        else:
            y1 = y0 + torch.einsum("b, s, sbf -> bf", dt_data, tableau.b, k)
        error_estimate = torch.einsum("b, s, sbf -> bf", dt_data, tableau.b_err, k)

        if tableau.fsal:
            state = ERKState(state.tableau, k[-1])

        return (
            StepResult(y1, error_estimate),
            ERKInterpolationData(tableau, t0, dt, y0, y1, k),
            state,
            None,
        )

    def build_interpolation(self, data: ERKInterpolationData) -> LocalInterpolation:
        raise NotImplementedError()



from typing import Any, Generic, NamedTuple, Optional, TypeVar

import torch.nn as nn

from ..interpolation import LocalInterpolation
from ..problems import InitialValueProblem
from ..terms import ODETerm
from ..typing import *


class StepResult(NamedTuple):
    y: DataTensor
    error_estimate: Optional[DataTensor]


MethodState = TypeVar("MethodState")
InterpolationData = TypeVar("InterpolationData")


class SingleStepMethod(nn.Module, Generic[MethodState, InterpolationData]):
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        f0: Optional[DataTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> MethodState:
        raise NotImplementedError()

    def step(
        self,
        term: Optional[ODETerm],
        running: AcceptTensor,
        y0: DataTensor,
        t0: TimeTensor,
        dt: TimeTensor,
        state: MethodState,
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[StepResult, InterpolationData, MethodState, Optional[StatusTensor]]:
        """Advance the solution from `y0` to `y0+dt`.

        Arguments
        ---------
        running
            Marks the instances in the batch that are actually still running. This is
            important for solvers with variable computation time such as implicit methods
            that use this information to short-circuit the evaluation of finished
            instances.
        y0
            Features at `t0`
        t0
            Initial point in time
        dt
            Step size of the step to make
        state
            Current state of the stepping method
        stats
            Tracked statistics for the current solve
        args
            Additional arguments for the ODE term

        Returns
        -------
        result
            Features `y1` at `t1 = t0 + dt` with an error estimate
        interpolation_data
            Additional information for interpolation between `t0` and `t1`
        state
            Updated state of the stepping method
        status
            Status to signify if integration should be stopped early (or None for
            all successes)
        """
        raise NotImplementedError()

    def merge_states(
        self, accept: AcceptTensor, current: MethodState, previous: MethodState
    ) -> MethodState:
        raise NotImplementedError()

    def build_interpolation(self, data: InterpolationData) -> LocalInterpolation:
        raise NotImplementedError()

    def convergence_order(self) -> int:
        raise NotImplementedError()



from typing import Any, Callable, Optional, Union

import torch

from .adjoints import AutoDiffAdjoint
from .problems import InitialValueProblem
from .single_step_methods import SingleStepMethod
from .solution import Solution
from .step_size_controllers import (
    FixedStepController,
    PIDController,
    StepSizeController,
)
from .terms import ODETerm
from .typing import *

METHODS = {}


def register_method(name: str, constructor: Callable[[ODETerm], SingleStepMethod]):
    METHODS[name] = constructor


def solve_ivp(
    f: Union[ODETerm, Callable[[TimeTensor, DataTensor], DataTensor]],
    y0: DataTensor,
    t_eval: EvaluationTimesTensor,
    *,
    t_span: Optional[tuple[TimeTensor, TimeTensor]] = None,
    method: Union[str, SingleStepMethod] = "tsit5",
    max_steps: Optional[int] = None,
    controller: Optional[StepSizeController] = None,
    dt0: Optional[TimeTensor] = None,
    args: Any = None,
) -> Solution:
    """Solve an initial value problem

    Arguments
    =========
    f
        The dynamics to solve
    y0
        Initial conditions
    t_eval
        Time points to evaluate the solution at
    t_span
        Start and end times of the integration. By default, integrate from the first to
        the last evaluation point.
    method
        Either the name of a registered stepping method, e.g. `"tsit5"`, or a stepping
        method object
    max_steps
        Stop the solver after this many steps
    controller
        Step size controller for the integration. By default a PID controller with
        `atol=1e-7, rtol=1e-7, pcoeff=0.2, icoeff=0.5, dcoeff=0.0` will be
        constructed.
    dt0
        An optional initial time step
    args
        Extra arguments to be passed to the integration term
    """

    if isinstance(f, ODETerm):
        term = f
    else:
        if args is not None:
            term = ODETerm(f, with_args=True)
        else:
            term = ODETerm(f)

    # TODO: Automatically reshape y0 into [batch, features] and back into its original
    # shape

    if not isinstance(method, SingleStepMethod):
        method = METHODS[method](term=term)

    if controller is None:
        controller = PIDController(
            term=term, atol=1e-7, rtol=1e-7, pcoeff=0.2, icoeff=0.5, dcoeff=0.0
        )

    adjoint = AutoDiffAdjoint(method, controller, max_steps=max_steps)

    batch_size = y0.shape[0]
    if t_eval is not None and t_eval.ndim == 1:
        t_eval = t_eval.expand((batch_size, -1))
    if t_span is not None:
        t_start, t_end = t_span
    else:
        t_start, t_end = t_eval[:, 0], t_eval[:, -1]
    if t_start.ndim == 0:
        t_start = t_start.expand(batch_size)
    if t_end.ndim == 0:
        t_end = t_end.expand(batch_size)
    problem = InitialValueProblem(y0, t_start, t_end, t_eval)
    return adjoint.solve(problem, term, dt0=dt0, args=args)



from typing import Optional

import torch

from ..interpolation import ThirdOrderPolynomialInterpolation
from ..terms import ODETerm
from .runge_kutta import ButcherTableau, ERKInterpolationData, ExplicitRungeKutta


class Heun(ExplicitRungeKutta):
    TABLEAU = ButcherTableau.from_lists(
        c=[0.0, 1.0], a=[[], [1.0]], b=[1 / 2, 1 / 2], b_low_order=[1.0, 0.0]
    )

    def __init__(self, term: Optional[ODETerm] = None):
        super().__init__(term, Heun.TABLEAU)

    @torch.jit.export
    def convergence_order(self):
        return 2

    @torch.jit.export
    def build_interpolation(self, data: ERKInterpolationData):
        return ThirdOrderPolynomialInterpolation.from_k(
            data.t0, data.dt, data.y0, data.y1, data.k
        )



from typing import Any, NamedTuple, Optional

import torch
import torch.nn as nn

from ..interpolation import LinearInterpolation
from ..problems import InitialValueProblem
from ..status_codes import Status
from ..terms import ODETerm
from ..typing import *
from .base import StepResult

NoneType = type(None)


class LinearInterpolationData(NamedTuple):
    t0: TimeTensor
    dt: TimeTensor
    y0: DataTensor
    y1: DataTensor


class Euler(nn.Module):
    def __init__(self, term: Optional[ODETerm]):
        super().__init__()

        self.term = term

    @torch.jit.export
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        f0: Optional[DataTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> NoneType:
        return None

    @torch.jit.export
    def step(
        self,
        term: Optional[ODETerm],
        running: AcceptTensor,
        y0: DataTensor,
        t0: TimeTensor,
        dt: TimeTensor,
        state: NoneType,
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[StepResult, LinearInterpolationData, NoneType, Optional[StatusTensor]]:
        term_ = term
        if torch.jit.is_scripting() or term_ is None:
            assert term is None, "The integration term is fixed for JIT compilation"
            term_ = self.term
        assert term_ is not None

        # Convert dt into the data dtype for dtype stability
        dt_data = dt.to(dtype=y0.dtype)

        y1 = torch.addcmul(y0, dt_data[:, None], term_.vf(t0, y0, stats, args))

        return (
            StepResult(y1, None),
            LinearInterpolationData(t0, dt, y0, y1),
            state,
            None,
        )

    @torch.jit.export
    def merge_states(
        self, accept: AcceptTensor, current: NoneType, previous: NoneType
    ) -> NoneType:
        return None

    @torch.jit.export
    def convergence_order(self):
        return 1

    @torch.jit.export
    def build_interpolation(self, data: LinearInterpolationData):
        return LinearInterpolation(data.t0, data.dt, data.y0, data.y1)



from enum import Enum


class Status(Enum):
    """A joint collection of all possible solver and step size controller status codes.

    Any status greater than 0, i.e. `Status.SUCCESS`, signifies some type of abnormal
    condition.

    This is an Enum instead of an IntEnum, because TorchScript only accepts Enums as of
    pytorch 1.11.
    """

    SUCCESS = 0
    GENERAL_ERROR = 1
    REACHED_DT_MIN = 2
    REACHED_MAX_STEPS = 3

    # The norm of the error ratio turned out infinite which points towards some bad
    # problems such as infinite or NaN y and the solving should be cancelled.
    INFINITE_NORM = 4



from typing import Optional

import sympy as sp
import torch

from ..interpolation import FourthOrderPolynomialInterpolation
from ..terms import ODETerm
from .runge_kutta import ButcherTableau, ERKInterpolationData, ExplicitRungeKutta


def compute_interpolation_weights():
    """Compute the interpolation weights for the Tsit5 interpolation coefficients of
    2nd, 3rd and 4th order.

    The original Tsit5 paper builds the 4th order interpolation polynomial as a linear
    combination of 7 polynomials. This function computes weights that give us the
    coefficients a, b, c in the standard polynomial form `a*x^4 + b*x^3 + ...` directly.
    This way, we can evaluate the interpolant more efficiently.
    """

    t, y0, dt, f1, f2, f3, f4, f5, f6, f7 = sp.symbols(
        "t y_0 dt f_1 f_2 f_3 f_4 f_5 f_6 f_7"
    )
    f = [f1, f2, f3, f4, f5, f6, f7]
    # fmt: off
    # The 7 basis functions of the interpolant
    b = [
        -1.0530884977290216 * t * (t - 1.3299890189751412) * (t**2 - 1.4364028541716351 * t + 0.7139816917074209),
        0.1017 * t**2 * (t**2 - 2.1966568338249754 * t + 1.2949852507374631),
        2.490627285651252793 * t**2 * (t**2 - 2.38535645472061657 * t + 1.57803468208092486),
        -16.54810288924490272 * (t - 1.21712927295533244) * (t - 0.61620406037800089) * t**2,
        47.37952196281928122 * (t - 1.203071208372362603) * (t - 0.658047292653547382) * t**2,
        -34.87065786149660974 * (t - 1.2) * (t - 0.666666666666666667) * t**2,
        2.5 * (t - 1) * (t - 0.6) * t**2
    ]
    # fmt: on
    interpolant = y0 + dt * sum(f_i * b_i for f_i, b_i in zip(f, b))
    # Fully expand the polynomial and collect the powers of t
    form = sp.collect(sp.expand(interpolant, t), t)
    # The coefficients of t^2, t^3 and t^4 are of the form
    #
    #     dt * \sum_i x_i f_i
    #
    # and we collect the x_i here in a matrix. Then the coefficients of the interpolant
    # can be found by a matrix multiplication between this and the k vector of the RK
    # stages.
    return [
        [float(form.coeff(t, i).coeff(f[j]).coeff(dt)) for j in range(len(f))]
        for i in range(2, 5)
    ]


class Tsit5(ExplicitRungeKutta):
    """The 5th order Runge-Kutta method by Tsitouras

    References
    ----------

    ```bibtex
    @article{tsitouras2011runge,
      title={Runge--Kutta pairs of order 5(4) satisfying only the first column
             simplifying assumption},
      author={Tsitouras, Charalampos},
      journal={Computers \\& Mathematics with Applications},
      volume={62},
      number={2},
      pages={770--775},
      year={2011},
      publisher={Elsevier}
    }
    ```
    """

    TABLEAU = ButcherTableau.from_lists(
        c=[0.0, 0.161, 0.327, 0.9, 0.9800255409045097, 1.0, 1.0],
        a=[
            # fmt: off
            [],
            [0.161],
            [-0.008480655492356989, 0.335480655492357],
            [2.8971530571054935, -6.359448489975075, 4.3622954328695815],
            [5.325864828439257, -11.748883564062828, 7.4955393428898365, -0.09249506636175525],
            [5.86145544294642, -12.92096931784711, 8.159367898576159, -0.071584973281401, -0.02826905039406838],
            [0.09646076681806523, 0.01, 0.4798896504144996, 1.379008574103742, -3.290069515436081, 2.324710524099774],
            # fmt: on
        ],
        b=[
            0.09646076681806523,
            0.01,
            0.4798896504144996,
            1.379008574103742,
            -3.290069515436081,
            2.324710524099774,
            0.0,
        ],
        # The paper introduces b-tilde as the weights of the lower-order interpolant but
        # the weights they give in the end are actually directly the weights for the
        # error estimate, see [1].
        #
        # [1] https://github.com/patrick-kidger/diffrax/issues/98
        b_err=[
            0.00178001105222577714,
            0.0008164344596567469,
            -0.007880878010261995,
            0.1447110071732629,
            -0.5823571654525552,
            0.45808210592918697,
            # The original paper has the sign of this coefficient wrong, see [1]
            #
            # [1] https://github.com/SciML/OrdinaryDiffEq.jl/issues/1654
            -1 / 66,
        ],
        b_other=compute_interpolation_weights(),
    )

    def __init__(self, term: Optional[ODETerm] = None):
        super().__init__(term, Tsit5.TABLEAU)

    @torch.jit.export
    def convergence_order(self):
        return 5

    @torch.jit.export
    def build_interpolation(self, data: ERKInterpolationData):
        y0 = data.y0
        dt = data.dt.to(dtype=y0.dtype)
        f0 = data.k[0]
        b = data.tableau.b_other
        assert b is not None

        B = torch.einsum("b, cs, sbf -> cbf", dt, b, data.k)
        c, b, a = B[0], B[1], B[2]
        d = dt[:, None] * f0
        e = y0

        coefficients = (e, d, c, b, a)
        return FourthOrderPolynomialInterpolation(
            data.t0, data.t0 + data.dt, coefficients
        )



from typing import Protocol

import torch

from .typing import *


def poly3eval(d, c, b, a, t, t0, t1):
    """Evaluate a 3rd order polynomial on the interval [t0, t1].

    The coefficients a..3 define the polynomial on the interval [0, 1].
    """
    x = ((t - t0) / (t1 - t0))[:, None].to(dtype=a.dtype)

    # Evaluate the polynomial with Horner's method
    y = a
    y = torch.addcmul(b, y, x)
    y = torch.addcmul(c, y, x)
    y = torch.addcmul(d, y, x)
    return y


def poly4eval(e, d, c, b, a, t, t0, t1):
    """Evaluate a 4th order polynomial on the interval [t0, t1].

    The coefficients a..e define the polynomial on the interval [0, 1].
    """
    x = ((t - t0) / (t1 - t0))[:, None].to(dtype=a.dtype)

    # Evaluate the polynomial with Horner's method
    y = a
    y = torch.addcmul(b, y, x)
    y = torch.addcmul(c, y, x)
    y = torch.addcmul(d, y, x)
    y = torch.addcmul(e, y, x)
    return y


class LocalInterpolation(Protocol):
    def evaluate(self, t: TimeTensor, idx: SampleIndexTensor) -> DataTensor:
        raise NotImplementedError()


class LinearInterpolation:
    def __init__(self, t0: TimeTensor, dt: TimeTensor, y0: DataTensor, y1: DataTensor):
        self.t0 = t0
        self.dt = dt
        self.y0 = y0
        self.dy = y1 - y0

    def evaluate(self, t: TimeTensor, idx: SampleIndexTensor) -> DataTensor:
        x = ((t - self.t0[idx]) / self.dt[idx])[:, None].to(dtype=self.y0.dtype)

        return torch.addcmul(self.y0[idx], self.dy[idx], x)


class ThirdOrderPolynomialInterpolation:
    """Third-order polynomial interpolation on [t0, t1].

    `coefficients` holds the coefficients of a third-order polynomial on [0, 1] in
    increasing order, i.e. `cofficients[i]` belongs to `x**i`.
    """

    def __init__(
        self,
        t0: TimeTensor,
        t1: TimeTensor,
        coefficients: tuple[DataTensor, DataTensor, DataTensor, DataTensor],
    ):
        self.t0 = t0
        self.t1 = t1
        self.coefficients = coefficients

    @staticmethod
    def from_k(
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        y1: DataTensor,
        k: torch.Tensor,
    ):
        """Find the coefficients from the k values of a Runge-Kutta step

        Computes the coefficients by fitting a third-order Hermite polynomial and then
        translating the found coefficients into the simple x**p polynomial form.
        """

        f0 = k[0]
        f1 = k[-1]
        dt_data = dt.to(dtype=y0.dtype)

        # The transformation from Hermite to normal coefficients is already included in
        # these formulas
        y0my1 = y0 - y1
        a = torch.add(dt_data[:, None] * (f0 + f1), y0my1, alpha=2)
        b = torch.add(dt_data[:, None] * torch.add(-f1, f0, alpha=-2), y0my1, alpha=-3)
        c = dt_data[:, None] * f0
        d = y0

        coefficients = (d, c, b, a)
        return ThirdOrderPolynomialInterpolation(t0, t0 + dt, coefficients)

    def evaluate(self, t: InterpTimeTensor, idx: SampleIndexTensor) -> InterpDataTensor:
        d, c, b, a = self.coefficients
        coeff = (d[idx], c[idx], b[idx], a[idx])
        return poly3eval(*coeff, t, self.t0[idx], self.t1[idx])

    def __repr__(self):
        return (
            f"ThirdOrderPolynomialInterpolation(t0={self.t0}, t1={self.t1}, "
            f"coefficients={self.coefficients})"
        )


class FourthOrderPolynomialInterpolation:
    """Polynomial interpolation on [t0, t1].

    `coefficients` holds the coefficients of a fourth-order polynomial on [0, 1] in
    increasing order, i.e. `cofficients[i]` belongs to `x**i`.
    """

    def __init__(
        self,
        t0: TimeTensor,
        t1: TimeTensor,
        coefficients: tuple[DataTensor, DataTensor, DataTensor, DataTensor, DataTensor],
    ):
        self.t0 = t0
        self.t1 = t1
        self.coefficients = coefficients

    @staticmethod
    def from_k(
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        y1: DataTensor,
        k: torch.Tensor,
        b_mid: torch.Tensor,
    ):
        dt_data = dt.to(dtype=y0.dtype)
        f0 = dt_data[:, None] * k[0]
        f1 = dt_data[:, None] * k[-1]
        y_mid = y0 + torch.einsum("b, s, sbf -> bf", dt_data, b_mid, k)

        a = (2 * (f1 - f0)).add(y1 + y0, alpha=-8).add(y_mid, alpha=16)
        b = (
            (5 * f0)
            .add(f1, alpha=-3)
            .add(y0, alpha=18)
            .add(y1, alpha=14)
            .add(y_mid, alpha=-32)
        )
        c = (
            f1.add(f0, alpha=-4)
            .add(y0, alpha=-11)
            .add(y1, alpha=-5)
            .add(y_mid, alpha=16)
        )
        d = f0
        e = y0

        coefficients = (e, d, c, b, a)
        return FourthOrderPolynomialInterpolation(t0, t0 + dt, coefficients)

    def evaluate(self, t: InterpTimeTensor, idx: SampleIndexTensor) -> InterpDataTensor:
        e, d, c, b, a = self.coefficients
        coeff = (e[idx], d[idx], c[idx], b[idx], a[idx])
        return poly4eval(*coeff, t, self.t0[idx], self.t1[idx])

    def __repr__(self):
        return (
            f"FourthOrderPolynomialInterpolation(t0={self.t0}, t1={self.t1}, "
            f"coefficients={self.coefficients})"
        )



from typing import Any, Callable, Final, Optional

import torch.nn as nn

from .problems import InitialValueProblem
from .typing import *


class ODETerm(nn.Module):
    with_args: Final[bool]
    with_stats: Final[bool]

    def __init__(
        self,
        f: Callable[[TimeTensor, DataTensor], DataTensor],
        *,
        with_stats: bool = True,
        with_args: bool = False,
    ):
        """Initialize an ODE term of the form `dy/dt = f(t, y)`.

        Arguments
        ---------
        f
            Right-hand side of the ODE
        with_stats
            If true, track statistics such as the number of function evaluations. If your
            dynamics are very fast to evaluate, disabling this can improve the performance
            of the solver by 1-2%.
        with_args
            If true, `f` will be passed an additional static argument in the third
            position.
        """

        super().__init__()

        self.f = f
        self.with_stats = with_stats
        self.with_args = with_args

    @torch.jit.export
    def init(self, problem: InitialValueProblem, stats: dict[str, Any]):
        if not self.with_stats:
            return
        # There is no reason for these to be on the GPU
        stats["n_f_evals"] = torch.zeros(
            problem.batch_size, device="cpu", dtype=torch.long
        )

    def vf(
        self, t: TimeTensor, y: DataTensor, stats: dict[str, Any], args: Any
    ) -> DataTensor:
        """Evaluate the vector field."""
        if self.with_stats:
            n_f_evals = stats["n_f_evals"]
            if torch.jit.is_scripting():
                assert isinstance(n_f_evals, torch.Tensor)
            n_f_evals.add_(1)

        if self.with_args:
            return self.f(t, y, args)
        else:
            return self.f(t, y)



from typing import Any

from .typing import *


class Solution:
    def __init__(
        self,
        ts: EvaluationTimesTensor,
        ys: SolutionDataTensor,
        stats: dict[str, Any],
        status: StatusTensor,
    ):
        self.ts = ts
        self.ys = ys
        self.stats = stats
        self.status = status

    def __repr__(self):
        return (
            f"Solution(ts={self.ts}, ys={self.ys}, "
            f"stats={self.stats}, status={self.status})"
        )



from math import sqrt
from typing import Any, Callable, Generic, NamedTuple, Optional, TypeVar

import torch
import torch.nn as nn

from .problems import InitialValueProblem
from .single_step_methods import StepResult
from .status_codes import Status
from .terms import ODETerm
from .typing import *

ControllerState = TypeVar("ControllerState")


class StepSizeController(nn.Module, Generic[ControllerState]):
    """A step size controller determines the size of integration steps."""

    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        method_order: int,
        dt0: Optional[TimeTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[TimeTensor, ControllerState, Optional[DataTensor]]:
        """Find the initial step size and initialize the controller state

        If the user suggests an initial step size, the controller should go with that
        one.

        If the controller evaluates the vector field at the initial step, for example
        to determine the initial step size, the evaluation at `t0` can be returned to
        save an evaluation in FSAL step methods.

        Arguments
        ---------
        term
            The integration term
        problem
            The problem to solve
        method_order
            Convergence order of the stepping method
        dt0
            An initial step size suggested by the user
        stats
            Container that the controller can initialize new statistics to track in
        args
            Static arguments for calls to the ODE term
        """
        raise NotImplementedError()

    def adapt_step_size(
        self,
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        step_result: StepResult,
        state: ControllerState,
        stats: dict[str, Any],
    ) -> tuple[AcceptTensor, TimeTensor, ControllerState, Optional[StatusTensor]]:
        """Adapt the integration step size based on the step just taken

        Arguments
        ---------
        t0
            Start time of the step
        dt
            Current step size
        y0
            State before the previous step
        step_result
            Result of the step just taken
        state
            Current controller state
        stats
            Tracked statistics for the current solve which can be updated in-place

        Returns
        -------
        accept
            Should the step be accepted or rejected?
        dt
            Next step size, either for the next step or to retry the current step
            if it was rejected
        state
            Next controller state
        status
            Status to signify if integration should be stopped early (or None for
            all sucesses)
        """
        raise NotImplementedError()

    def merge_states(
        self, running: AcceptTensor, current: ControllerState, previous: ControllerState
    ) -> ControllerState:
        """Merge two controller states

        Any batch-specific state should be updated so that it updates only for the still
        running instances and stays constant for finished instances.

        Arguments
        ---------
        running
            Marks the instances in the batch that are still being solved
        current
            The controller state at the end of the current iteration
        previous
            The previous controller state
        """
        raise NotImplementedError()


class FixedStepState(NamedTuple):
    accept_all: AcceptTensor
    dt0: TimeTensor


class FixedStepController(StepSizeController[FixedStepState]):
    """A fixed-step step size controller.

    Does not actually control anything. Just accepts any result and keeps the step size
    fixed.
    """

    @torch.jit.export
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        method_order: int,
        dt0: Optional[TimeTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ):
        assert dt0 is not None, "Fixed step size solving requires you to configure dt0"
        return (
            dt0,
            FixedStepState(
                accept_all=torch.ones(
                    problem.batch_size, device=problem.device, dtype=torch.bool
                ),
                dt0=dt0,
            ),
            None,
        )

    @torch.jit.export
    def adapt_step_size(
        self,
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        step_result: StepResult,
        state: FixedStepState,
        stats: dict[str, Any],
    ) -> tuple[AcceptTensor, TimeTensor, FixedStepState, Optional[StatusTensor]]:
        return state.accept_all, state.dt0, state, None

    @torch.jit.export
    def merge_states(
        self, running: AcceptTensor, current: FixedStepState, previous: FixedStepState
    ) -> FixedStepState:
        return current


def rms_norm(y: DataTensor) -> NormTensor:
    """Root mean squared error norm.

    As suggested in [1], Equation (4.11).

    References
    ----------
    .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
           Equations I: Nonstiff Problems", Sec. II.4, 2nd edition.
    """
    # `vector_norm` autmatically deals with complex vectors correctly
    return torch.linalg.vector_norm(y / sqrt(y.shape[1]), ord=2, dim=1)


def max_norm(y: DataTensor) -> NormTensor:
    """Maximums norm."""
    return torch.linalg.vector_norm(y, dim=1, ord=torch.inf)


class AdaptiveStepSizeController(
    StepSizeController[ControllerState], Generic[ControllerState]
):
    def initial_state(
        self,
        method_order: int,
        problem: InitialValueProblem,
        dt_min: Optional[TimeTensor],
        dt_max: Optional[TimeTensor],
    ) -> ControllerState:
        raise NotImplementedError()

    def update_state(
        self,
        state: ControllerState,
        y0: DataTensor,
        dt: TimeTensor,
        error_ratio: Optional[NormTensor],
        accept: Optional[AcceptTensor],
    ) -> ControllerState:
        raise NotImplementedError()

    def dt_factor(self, state: ControllerState, error_ratio: NormTensor):
        raise NotImplementedError()


class IntegralState:
    def __init__(
        self,
        method_order: int,
        almost_zero: torch.Tensor,
        dt_min: Optional[torch.Tensor] = None,
        dt_max: Optional[torch.Tensor] = None,
    ):
        self.method_order = method_order
        self.almost_zero = almost_zero
        self.dt_min = dt_min
        self.dt_max = dt_max

    def update_error_ratios(
        self, prev_error_ratio: NormTensor, prev_prev_error_ratio: NormTensor
    ):
        return self

    def __repr__(self):
        return (
            f"IState(method_order={self.method_order}, "
            f"almost_zero={self.almost_zero}, "
            f"dt_min={self.dt_min}), "
            f"dt_max={self.dt_max})"
        )

    @staticmethod
    def default(
        *,
        method_order: int,
        batch_size: int,
        dtype: torch.dtype,
        device: Optional[torch.device],
        dt_min: Optional[torch.Tensor],
        dt_max: Optional[torch.Tensor],
    ):
        # Pre-allocate a fixed, very small number as a lower bound for the error ratio
        if dtype == torch.float16:
            float_min = 1e-5
        else:
            float_min = 1e-38
        almost_zero = torch.tensor(float_min, dtype=dtype, device=device)
        return IntegralState(method_order, almost_zero, dt_min, dt_max)


class IntegralController(nn.Module):
    """The simplest controller that scales the step size proportional to the error."""

    def __init__(
        self,
        atol: float,
        rtol: float,
        *,
        term: Optional[ODETerm] = None,
        norm: Callable[[DataTensor], NormTensor] = rms_norm,
        dt_min: Optional[float] = None,
        dt_max: Optional[float] = None,
        safety: float = 0.9,
        factor_min: float = 0.2,
        factor_max: float = 10.0,
    ):
        super().__init__()

        self.register_buffer("atol", torch.tensor(atol))
        self.register_buffer("rtol", torch.tensor(rtol))
        self.term = term
        self.norm = norm
        self.dt_min = dt_min
        self.dt_max = dt_max

        self.safety = safety
        self.factor_min = factor_min
        self.factor_max = factor_max

    def dt_factor(self, state: IntegralState, error_ratio: NormTensor):
        """Compute the growth factor of the timestep."""

        k_I = 1.0 / state.method_order
        factor = self.safety * error_ratio ** (-k_I)
        return torch.clamp(factor, min=self.factor_min, max=self.factor_max)

    def initial_state(
        self,
        method_order: int,
        problem: InitialValueProblem,
        dt_min: Optional[TimeTensor],
        dt_max: Optional[TimeTensor],
    ) -> IntegralState:
        return IntegralState.default(
            method_order=method_order,
            batch_size=problem.batch_size,
            dtype=problem.data_dtype,
            device=problem.device,
            dt_min=dt_min,
            dt_max=dt_max,
        )

    @torch.jit.export
    def merge_states(
        self, running: AcceptTensor, current: IntegralState, previous: IntegralState
    ) -> IntegralState:
        return current

    def update_state(
        self,
        state: IntegralState,
        y0: DataTensor,
        dt: TimeTensor,
        error_ratio: Optional[NormTensor],
        accept: Optional[AcceptTensor],
    ) -> IntegralState:
        return state

    ################################################################################
    # The following methods should be on AdaptiveStepSizeController if TorchScript #
    # supports inheritance at some point                                           #
    ################################################################################

    @torch.jit.export
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        method_order: int,
        dt0: Optional[TimeTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[TimeTensor, IntegralState, Optional[DataTensor]]:
        if dt0 is None:
            dt0, f0 = self._select_initial_step(
                term,
                problem.t_start,
                problem.y0,
                problem.time_direction,
                method_order,
                stats,
                args,
            )
        else:
            f0 = None
        dt_min = self.dt_min
        if dt_min is not None:
            dt_min = torch.tensor(
                dt_min, dtype=problem.time_dtype, device=problem.device
            )
        dt_max = self.dt_max
        if dt_max is not None:
            dt_max = torch.tensor(
                dt_max, dtype=problem.time_dtype, device=problem.device
            )
        return dt0, self.initial_state(method_order, problem, dt_min, dt_max), f0

    @torch.jit.export
    def adapt_step_size(
        self,
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        step_result: StepResult,
        state: IntegralState,
        stats: dict[str, Any],
    ) -> tuple[AcceptTensor, TimeTensor, IntegralState, Optional[StatusTensor]]:
        y1, error_estimate = step_result.y, step_result.error_estimate

        if error_estimate is None:
            # If the stepping method could not provide an error estimate, we interpret
            # this as an error estimate that gets the step accepted without changing the
            # step size, i.e. as an error ratio of 1 (disregarding the safety factor).
            return (
                torch.ones_like(dt, dtype=torch.bool),
                dt,
                self.update_state(state, y0, dt, None, None),
                None,
            )

        # Compute error ratio and decide on step acceptance
        error_bounds = torch.add(
            self.atol, torch.maximum(y0.abs(), y1.abs()), alpha=self.rtol
        )
        error = error_estimate.abs()
        # We lower-bound the error ratio by some small number to avoid division by 0 in
        # `dt_factor`.
        error_ratio = torch.maximum(self.norm(error / error_bounds), state.almost_zero)
        accept = error_ratio < 1.0

        # Adapt the step size
        dt_next = dt * self.dt_factor(state, error_ratio).to(dtype=dt.dtype)

        # Check for infinities and NaN
        status = torch.where(
            torch.isfinite(error_ratio),
            Status.SUCCESS.value,
            Status.INFINITE_NORM.value,
        )

        # Enforce the minimum and maximum step size
        dt_min = state.dt_min
        dt_max = state.dt_max
        if dt_min is not None or dt_max is not None:
            abs_dt_next = dt_next.abs()
            dt_next = torch.sign(dt_next) * torch.clamp(abs_dt_next, dt_min, dt_max)
            if dt_min is not None:
                status = torch.where(
                    abs_dt_next < dt_min, Status.REACHED_DT_MIN.value, status
                )

        return (
            accept,
            dt_next,
            self.update_state(state, y0, dt, error_ratio, accept),
            status,
        )

    def _select_initial_step(
        self,
        term: Optional[ODETerm],
        t0: TimeTensor,
        y0: DataTensor,
        direction: torch.Tensor,
        convergence_order: int,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[TimeTensor, DataTensor]:
        """Empirically select a good initial step.

        This is an adaptation of the algorithm described in [1]_. We changed it in such a
        way that the tolerances apply to the norms instead of the components of `y`.

        References
        ----------
        .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential Equations
        I: Nonstiff Problems", Sec. II.4, 2nd edition.
        """

        if torch.jit.is_scripting() or term is None:
            assert term is None, "The integration term is fixed for JIT compilation"
            term = self.term
        assert term is not None

        norm = self.norm
        f0 = term.vf(t0, y0, stats, args)

        error_bounds = torch.add(self.atol, torch.abs(y0), alpha=self.rtol)
        inv_scale = torch.reciprocal(error_bounds)

        d0 = norm(y0 * inv_scale)
        d1 = norm(f0 * inv_scale)

        small_number = torch.tensor(1e-6, dtype=d0.dtype, device=d0.device)
        dt0 = torch.where((d0 < 1e-5) | (d1 < 1e-5), small_number, 0.01 * d0 / d1)

        y1 = torch.addcmul(y0, (direction.to(dtype=dt0.dtype) * dt0)[:, None], f0)
        f1 = term.vf(
            torch.addcmul(t0, direction.to(dtype=t0.dtype), dt0.to(dtype=t0.dtype)),
            y1,
            stats,
            args,
        )

        d2 = norm((f1 - f0) * inv_scale) / dt0

        maxd1d2 = torch.maximum(d1, d2)
        dt1 = torch.where(
            maxd1d2 <= 1e-15,
            torch.maximum(small_number, dt0 * 1e-3),
            (0.01 / maxd1d2) ** (1.0 / convergence_order),
        )

        return (direction * torch.minimum(100 * dt0, dt1)).to(dtype=t0.dtype), f0


class PIDState:
    def __init__(
        self,
        method_order: int,
        prev_error_ratio: NormTensor,
        prev_prev_error_ratio: NormTensor,
        almost_zero: torch.Tensor,
        dt_min: Optional[torch.Tensor] = None,
        dt_max: Optional[torch.Tensor] = None,
    ):
        self.method_order = method_order
        self.prev_error_ratio = prev_error_ratio
        self.prev_prev_error_ratio = prev_prev_error_ratio
        self.almost_zero = almost_zero
        self.dt_min = dt_min
        self.dt_max = dt_max

    def update_error_ratios(
        self, prev_error_ratio: NormTensor, prev_prev_error_ratio: NormTensor
    ):
        return PIDState(
            self.method_order,
            prev_error_ratio,
            prev_prev_error_ratio,
            self.almost_zero,
            self.dt_min,
            self.dt_max,
        )

    def __repr__(self):
        return (
            f"PIDState(method_order={self.method_order}, "
            f"prev_error_ratio={self.prev_error_ratio}), "
            f"prev_prev_error_ratio={self.prev_prev_error_ratio}, "
            f"almost_zero={self.almost_zero}, "
            f"dt_min={self.dt_min}), "
            f"dt_max={self.dt_max})"
        )

    @staticmethod
    def default(
        *,
        method_order: int,
        batch_size: int,
        dtype: torch.dtype,
        device: Optional[torch.device],
        dt_min: Optional[torch.Tensor],
        dt_max: Optional[torch.Tensor],
    ):
        default_ratio = torch.ones(batch_size, dtype=dtype, device=device)
        # Pre-allocate a fixed, very small number as a lower bound for the error ratio
        if dtype == torch.float16:
            float_min = 1e-5
        else:
            float_min = 1e-38
        almost_zero = torch.tensor(float_min, dtype=dtype, device=device)
        return PIDState(
            method_order, default_ratio, default_ratio, almost_zero, dt_min, dt_max
        )


class PIDController(nn.Module):
    """A PID step size controller.

    The formula for the dt scaling factor with PID control is taken from [1], Equation
    (34).

    References
    ----------
    [1] Söderlind, G. (2003). Digital Filters in Adaptive Time-Stepping. ACM
        Transactions on Mathematical Software, 29, 1–26.
    """

    def __init__(
        self,
        atol: float,
        rtol: float,
        pcoeff: float,
        icoeff: float,
        dcoeff: float,
        *,
        term: Optional[ODETerm] = None,
        norm: Callable[[DataTensor], NormTensor] = rms_norm,
        dt_min: Optional[float] = None,
        dt_max: Optional[float] = None,
        safety: float = 0.9,
        factor_min: float = 0.2,
        factor_max: float = 10.0,
    ):
        super().__init__()

        self.register_buffer("atol", torch.tensor(atol))
        self.register_buffer("rtol", torch.tensor(rtol))
        self.term = term
        self.norm = norm
        self.dt_min = dt_min
        self.dt_max = dt_max

        self.pcoeff = pcoeff
        self.icoeff = icoeff
        self.dcoeff = dcoeff
        self.safety = safety
        self.factor_min = factor_min
        self.factor_max = factor_max

    def dt_factor(self, state: PIDState, error_ratio: NormTensor):
        """Compute the growth factor of the timestep."""

        # This is an instantiation of Equation (34) in the Söderlind paper where we have
        # factored out the safety coefficient. I have not found a reference for dividing
        # the PID coefficients by the order of the solver but DifferentialEquations.jl
        # and diffrax both do it, so we do it too. Note that our error ratio is the
        # reciprocal of Söderlind's error ratio (except for the safety factor).
        # Therefore, the factor exponents have the opposite sign from the paper.
        #
        # Interesting thing from the introduction of that paper is that you work with p
        # if you want per-step-error-control and p+1 if you want
        # per-unit-step-error-control where p is the convergence order of the stepping
        # method.
        order = state.method_order
        k_I, k_P, k_D = self.icoeff / order, self.pcoeff / order, self.dcoeff / order

        factor1 = error_ratio ** (-(k_I + k_P + k_D))
        factor2 = state.prev_error_ratio ** (k_P + 2 * k_D)
        factor3 = state.prev_prev_error_ratio**-k_D
        factor = self.safety * factor1 * factor2 * factor3

        return torch.clamp(factor, min=self.factor_min, max=self.factor_max)

    def initial_state(
        self,
        method_order: int,
        problem: InitialValueProblem,
        dt_min: Optional[TimeTensor],
        dt_max: Optional[TimeTensor],
    ) -> PIDState:
        return PIDState.default(
            method_order=method_order,
            batch_size=problem.batch_size,
            dtype=problem.data_dtype,
            device=problem.device,
            dt_min=dt_min,
            dt_max=dt_max,
        )

    @torch.jit.export
    def merge_states(
        self, running: AcceptTensor, current: PIDState, previous: PIDState
    ) -> PIDState:
        return current.update_error_ratios(
            torch.where(running, current.prev_error_ratio, previous.prev_error_ratio),
            torch.where(
                running, current.prev_prev_error_ratio, previous.prev_prev_error_ratio
            ),
        )

    def update_state(
        self,
        state: PIDState,
        y0: DataTensor,
        dt: TimeTensor,
        error_ratio: Optional[NormTensor],
        accept: Optional[AcceptTensor],
    ) -> PIDState:
        if error_ratio is None:
            return state.update_error_ratios(
                prev_error_ratio=y0.new_ones(dt.shape),
                prev_prev_error_ratio=state.prev_error_ratio,
            )
        else:
            assert accept is not None
            return state.update_error_ratios(
                prev_error_ratio=torch.where(
                    accept, error_ratio, state.prev_error_ratio
                ),
                prev_prev_error_ratio=torch.where(
                    accept, state.prev_error_ratio, state.prev_prev_error_ratio
                ),
            )

    ################################################################################
    # The following methods should be on AdaptiveStepSizeController if TorchScript #
    # supports inheritance at some point                                           #
    ################################################################################

    @torch.jit.export
    def init(
        self,
        term: Optional[ODETerm],
        problem: InitialValueProblem,
        method_order: int,
        dt0: Optional[TimeTensor],
        *,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[TimeTensor, PIDState, Optional[DataTensor]]:
        if dt0 is None:
            dt0, f0 = self._select_initial_step(
                term,
                problem.t_start,
                problem.y0,
                problem.time_direction,
                method_order,
                stats,
                args,
            )
        else:
            f0 = None
        dt_min = self.dt_min
        if dt_min is not None:
            dt_min = torch.tensor(
                dt_min, dtype=problem.time_dtype, device=problem.device
            )
        dt_max = self.dt_max
        if dt_max is not None:
            dt_max = torch.tensor(
                dt_max, dtype=problem.time_dtype, device=problem.device
            )
        return dt0, self.initial_state(method_order, problem, dt_min, dt_max), f0

    @torch.jit.export
    def adapt_step_size(
        self,
        t0: TimeTensor,
        dt: TimeTensor,
        y0: DataTensor,
        step_result: StepResult,
        state: PIDState,
        stats: dict[str, Any],
    ) -> tuple[AcceptTensor, TimeTensor, PIDState, Optional[StatusTensor]]:
        y1, error_estimate = step_result.y, step_result.error_estimate

        if error_estimate is None:
            # If the stepping method could not provide an error estimate, we interpret
            # this as an error estimate that gets the step accepted without changing the
            # step size, i.e. as an error ratio of 1 (disregarding the safety factor).
            return (
                torch.ones_like(dt, dtype=torch.bool),
                dt,
                self.update_state(state, y0, dt, None, None),
                None,
            )

        # Compute error ratio and decide on step acceptance
        error_bounds = torch.add(
            self.atol, torch.maximum(y0.abs(), y1.abs()), alpha=self.rtol
        )
        error = error_estimate.abs()
        # We lower-bound the error ratio by some small number to avoid division by 0 in
        # `dt_factor`.
        error_ratio = torch.maximum(self.norm(error / error_bounds), state.almost_zero)
        accept = error_ratio < 1.0

        # Adapt the step size
        dt_next = dt * self.dt_factor(state, error_ratio).to(dtype=dt.dtype)

        # Check for infinities and NaN
        status = torch.where(
            torch.isfinite(error_ratio),
            Status.SUCCESS.value,
            Status.INFINITE_NORM.value,
        )

        # Enforce the minimum and maximum step size
        dt_min = state.dt_min
        dt_max = state.dt_max
        if dt_min is not None or dt_max is not None:
            abs_dt_next = dt_next.abs()
            dt_next = torch.sign(dt_next) * torch.clamp(abs_dt_next, dt_min, dt_max)
            if dt_min is not None:
                status = torch.where(
                    abs_dt_next < dt_min, Status.REACHED_DT_MIN.value, status
                )

        return (
            accept,
            dt_next,
            self.update_state(state, y0, dt, error_ratio, accept),
            status,
        )

    def _select_initial_step(
        self,
        term: Optional[ODETerm],
        t0: TimeTensor,
        y0: DataTensor,
        direction: torch.Tensor,
        convergence_order: int,
        stats: dict[str, Any],
        args: Any,
    ) -> tuple[TimeTensor, DataTensor]:
        """Empirically select a good initial step.

        This is an adaptation of the algorithm described in [1]_. We changed it in such a
        way that the tolerances apply to the norms instead of the components of `y`.

        References
        ----------
        .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential Equations
        I: Nonstiff Problems", Sec. II.4, 2nd edition.
        """

        if torch.jit.is_scripting() or term is None:
            assert term is None, "The integration term is fixed for JIT compilation"
            term = self.term
        assert term is not None

        norm = self.norm
        f0 = term.vf(t0, y0, stats, args)

        error_bounds = torch.add(self.atol, torch.abs(y0), alpha=self.rtol)
        inv_scale = torch.reciprocal(error_bounds)

        d0 = norm(y0 * inv_scale)
        d1 = norm(f0 * inv_scale)

        small_number = torch.tensor(1e-6, dtype=d0.dtype, device=d0.device)
        dt0 = torch.where((d0 < 1e-5) | (d1 < 1e-5), small_number, 0.01 * d0 / d1)

        y1 = torch.addcmul(y0, (direction.to(dtype=dt0.dtype) * dt0)[:, None], f0)
        f1 = term.vf(
            torch.addcmul(t0, direction.to(dtype=t0.dtype), dt0.to(dtype=t0.dtype)),
            y1,
            stats,
            args,
        )

        d2 = norm((f1 - f0) * inv_scale) / dt0

        maxd1d2 = torch.maximum(d1, d2)
        dt1 = torch.where(
            maxd1d2 <= 1e-15,
            torch.maximum(small_number, dt0 * 1e-3),
            (0.01 / maxd1d2) ** (1.0 / convergence_order),
        )

        return (direction * torch.minimum(100 * dt0, dt1)).to(dtype=t0.dtype), f0
